import re

def extract_number_expression(name):
    expr = ""
    is_num = False
    nums = 0
    for c in name:
        if c.isnumeric():
            is_num = True
        else:
            if is_num:
                is_num = False
                expr += "([0-9]+)"
                nums += 1
            expr += re.escape(c)
    if is_num:
        expr += "([0-9]+)"
        nums += 1
    expr = "^" + expr + "$"
    return expr, nums


from .base import hint_matcher

@hint_matcher
def match_passed_hints(state, match):
    def is_match(out_node, in_node, hint):
        return (hint[0] in out_node.full_prefix) == (hint[1] in in_node.full_prefix)

    return state.hints, is_match, lambda hint: f"passed hints"


from .base import hint_matcher

@hint_matcher
def match_by_paired_predecessors(state, match):
    def get_predecessors(node, nodes):
        parents = [node] if any(node.is_predecessor_of(n) for n in nodes) else []
        if not node.parent is None:
            parents = get_predecessors(node.parent, nodes) + parents
        return parents

    hints = set()
    for out_node, in_node in state.paired_nodes:
        out_predecessors = get_predecessors(out_node, match.out_nodes)
        in_predecessors = get_predecessors(in_node, match.in_nodes)
        for out_predecessor in out_predecessors:
            for in_predecessor in in_predecessors:
                hints.add((out_predecessor, in_predecessor))
    hints = sorted(hints, key=lambda t: -(t[0].depth + t[1].depth))

    def is_match(out_node, in_node, hint):
        return hint[0].is_predecessor_of(out_node) == hint[1].is_predecessor_of(in_node)

    return hints, is_match, lambda hint: f"paired predecessor OUT {hint[0].full_prefix} IN {hint[1].full_prefix}"


from . import match_by_structured_shapes
from . import match_by_paired_parents
from . import match_by_paired_children
from . import match_unique_leafs
from . import match_number_regex
from . import match_known_paired_number_regex
from . import match_equivalent_hardcoded_leafs
from . import match_passed_hints
from . import match_by_paired_predecessors
from . import match_by_paired_prefixes


from .base import hint_matcher

@hint_matcher
def match_by_paired_prefixes(state, match):
    def get_prefixes(name, separator, nodes):
        prefixes = []
        prefix = ""
        for t in name.split(separator):
            prefix = (prefix + separator + t) if len(prefix) > 0 else t
            prefixes.append(prefix)
        def check(prefix):
            starting_with = list(n.full_prefix.startswith(prefix) for n in nodes)
            return any(starting_with)
        prefixes = [p for p in prefixes if check(p)]
        return prefixes

    hints = set()
    for out_node, in_node in state.paired_nodes:
        out_prefixes = get_prefixes(out_node.full_prefix, state.out_separator, match.out_nodes)
        in_prefixes = get_prefixes(in_node.full_prefix, state.in_separator, match.in_nodes)
        for out_prefix in out_prefixes:
            for in_prefix in in_prefixes:
                hints.add((out_prefix, in_prefix))
    hints = sorted(hints, key=lambda t: -(len(t[0]) + len(t[1])))

    def is_match(out_node, in_node, hint):
        return out_node.full_prefix.startswith(hint[0]) == in_node.full_prefix.startswith(hint[1])

    return hints, is_match, lambda hint: f"paired prefix hint OUT {hint[0]} IN  {hint[1]}"



import numpy as np
import time, types
from collections import defaultdict
from . import tree
from .matchers.base import State, Match
from . import matchers
from functools import partial

LOAD_PREFIX = "load_prefix"

possible_separators = ["/", "."]
def get_separator_for(names, type):
    has_separator = [any(s in name for name in names) for s in possible_separators]
    n = np.count_nonzero(has_separator)
    if n == 0:
        return None
    elif n == 1:
        return possible_separators[np.argmax(has_separator, axis=0)]
    else:
        raise ValueError(f"Could not implicitly determine separator in weight names. Please provide argument {type}.")



def adapt(in_values, out_values, in_format=None, out_format=None, in_separator=None, out_separator=None, hints=[], verbose=False):
    t0 = time.time()
    if isinstance(out_values, dict):
        single_input = True
        out_values = [out_values]
    if isinstance(in_values, dict):
        in_values = [in_values]

    # Get separator in input values
    if in_separator is None:
        keys = set()
        def recurse(d):
            if isinstance(d, dict):
                for k, v in d.items():
                    keys.add(k)
                    recurse(v)
        for d in in_values:
            recurse(d)
        in_separator = get_separator_for(keys, "in_separator")
        if in_separator is None:
            in_separator = possible_separators[0]

    # Get separator in output values
    if out_separator is None:
        keys = set()
        def recurse(d):
            if isinstance(d, dict):
                for k, v in d.items():
                    keys.add(k)
                    recurse(v)
        for d in out_values:
            recurse(d)
        out_separator = get_separator_for(keys, "out_separator")
        if out_separator is None:
            out_separator = possible_separators[0]

    # Flatten input values
    flat_in_values = {}
    def recurse(source, key=()):
        if isinstance(source, dict):
            for k, v in source.items():
                recurse(v, key + (k,))
        else:
            if len(key) == 0:
                raise ValueError("Input values must be a dictionary")
            name = in_separator.join(key)
            if name in flat_in_values:
                raise ValueError(f"Duplicate input name {name}")
            flat_in_values[name] = np.asarray(source)
    for d in in_values:
        recurse(d)
    in_values = flat_in_values

    # Flatten output values and remember tree structure
    flat_out_values = {}
    def recurse(source, key=()):
        if isinstance(source, dict):
            treedef = {}
            for k, v in source.items():
                treedef[k] = recurse(v, key + (k,))
            return treedef
        else:
            if len(key) == 0:
                raise ValueError("Output values must be a dictionary")
            name = out_separator.join(key)
            if name in flat_out_values:
                raise ValueError(f"Duplicate output name {name}")
            source = np.asarray(source)
            flat_out_values[name] = source
            return (name, source.shape, source.dtype)
    out_treedefs = []
    for d in out_values:
        out_treedefs.append(recurse(d))
    out_values = flat_out_values



    state = State(
        in_values,
        {LOAD_PREFIX + out_separator + k: v for k, v in out_values.items()},
        in_separator=in_separator,
        out_separator=out_separator,
        in_format=in_format,
        out_format=out_format,
        hints=hints,
        verbose=verbose,
    )

    # Build module trees
    out_tree = tree.build(state.out_values)
    in_tree = tree.build(state.in_values)

    # Initialize matches
    matches = [Match(list(in_tree), list(out_tree), "Root")]

    # Construct heuristics
    ops = [
        partial(matchers.match_by_structured_shapes.match_by_structured_shapes, use_product=True),
        matchers.match_by_paired_parents.match_by_paired_parents,
        matchers.match_by_paired_children.match_by_paired_children,
        matchers.match_unique_leafs.match_unique_leafs,
        matchers.match_number_regex.match_number_regex,
        matchers.match_known_paired_number_regex.match_known_paired_number_regex,
        matchers.match_equivalent_hardcoded_leafs.match_equivalent_hardcoded_leafs,
        matchers.match_passed_hints.match_passed_hints,
        matchers.match_by_paired_prefixes.match_by_paired_prefixes,
        matchers.match_by_paired_predecessors.match_by_paired_predecessors,
    ]
    ops = {op.__name__ if not isinstance(op, partial) else op.func.__name__: op for op in ops}

    # Run matching heuristics
    times = defaultdict(lambda: 0.0)
    changed = True
    while changed:
        changed = False
        for name, op in ops.items():
            if verbose:
                print(f"OP: Trying {name}")
            start = time.time()
            matches, changed = op(state, matches)

            times[name] += time.time() - start
            if changed:
                if verbose:
                    print(f"OP: Changed by {name}")
                break

    if verbose:
        print("Times (sec) per operation:")
        for k, v in sorted(times.items(), key=lambda t: t[1]):
            print(f"    {k} {v}")

    if len(matches) > 0:
        print()
        for match in matches:
            print("Failed to pair the following nodes")
            for n in match.out_nodes:
                print(f"    OUT {n.full_prefix} {n.get_structured_shapes()}")
            for n in match.in_nodes:
                print(f"    IN  {n.full_prefix} {n.get_structured_shapes()}")
        raise ValueError("Failed to pair input values with output values")

    if verbose:
        print()
        print("Paired values:")
        mapping = {out_node.full_prefix: in_node.full_prefix for out_node, in_node in state.paired_nodes if out_node.is_leaf() and in_node.is_leaf()}
        for out_value in state.out_values:
            print(f"    {out_value.name} -> {mapping[out_value.name]}")

    # Matching successful! Now build tree of output values
    out_values_by_name = {v.name: v for v in state.out_values}
    def recurse(x):
        if isinstance(x, dict):
            return {k: recurse(v) for k, v in x.items()}
        else:
            out_name, out_shape, dtype = x
            in_value = out_values_by_name[LOAD_PREFIX + out_separator + out_name].other.value
            in_name = out_values_by_name[LOAD_PREFIX + out_separator + out_name].other.name

            assert np.prod(in_value.shape) == np.prod(out_shape)

            out_value = state.adapt_format(in_value, out_shape, in_name, out_name)

            return out_value

    out_values = [recurse(treedef) for treedef in out_treedefs]

    if verbose:
        print(f"Matching took {time.time() - t0:.2f} seconds")

    return out_values[0] if single_input else out_values



import functools
import numpy as np

def get_common_prefix(strings):
    strings = list(strings)
    s = ""
    while all(len(s) < len(sn) for sn in strings) and len(set(sn[len(s)] for sn in strings)) == 1:
        s = s + strings[0][len(s)]
    return s

class Node:
    def __init__(self, value, relative_prefix, full_prefix, direct_children, depth):
        self.value = value
        self.relative_prefix = relative_prefix
        self.full_prefix = full_prefix
        self.direct_children = direct_children
        self.parent = None
        self.depth = depth
        self.other = None
        if self.is_leaf():
            assert not value is None

    def is_leaf(self):
        return len(self.direct_children) == 0

    def is_predecessor_of(self, other):
        return id(other) == id(self) or (not other.parent is None and self.is_predecessor_of(other.parent))

    def get_structured_shapes(self, use_product=False, ignore_shape_one=False, remove_trivial_nodes=False, flatten=False):
        """Returns an object representing the shapes of weights in this node.

        The returned object is invariant to the ordering of children.

        Args:
            use_product: Whether the object should be invariant to the specific shape of the weight and only represent the number of elements. Defaults to ``False``.
            ignore_shape_one: Whether the object should be invariant to existence of axes of size 1. Defaults to ``True``.
            remove_trivial_nodes: Whether the object should be invariant to the existence of intermediate nodes that have only a single child node. Defaults to ``False``.
            flatten: Whether the object should be invariant to nesting of children. Defaults to ``False``.

        Returns:
            An id object representing the shapes of weights in this node.
        """
        if self.value is None:
            value_shape = []
        else:
            if use_product:
                value_shape = [(int(np.prod(self.value.shape)),)]
            else:
                value_shape = [tuple(s for s in self.value.shape if not ignore_shape_one or s != 1)]
        direct_children = self.direct_children
        if remove_trivial_nodes:
            def drop(n):
                if len(n.direct_children) == 1:
                    return drop(n.direct_children[0])
                else:
                    return n
            direct_children = [drop(c) for c in direct_children]
        children_shapes = [c.get_structured_shapes(ignore_shape_one=ignore_shape_one, remove_trivial_nodes=remove_trivial_nodes, flatten=flatten) for c in direct_children]
        if flatten:
            def flatten(shapes):
                result = []
                for s in shapes:
                    if isinstance(s, tuple) and all(not isinstance(x, tuple) for x in s):
                        result.append(s)
                    else:
                        result.extend(flatten(s))
                return result
            children_shapes = flatten(children_shapes)
        def compare_int(i1, i2):
            if i1 < i2:
                return -1
            elif i1 > i2:
                return 1
            else:
                return 0
        def compare(shapes1, shapes2):
            if isinstance(shapes1, tuple) and isinstance(shapes2, tuple):
                c = compare_int(len(shapes1), len(shapes2))
                if c != 0:
                    return c
                for s1, s2 in zip(shapes1, shapes2):
                    c = compare(s1, s2)
                    if c != 0:
                        return c
                return 0
            elif isinstance(shapes1, int) and isinstance(shapes2, tuple):
                return 1
            elif isinstance(shapes1, tuple) and isinstance(shapes2, int):
                return -1
            else:
                return compare_int(shapes1, shapes2)
        return tuple(sorted(value_shape + children_shapes, key=functools.cmp_to_key(compare)))

    def __iter__(self):
        yield self
        for c in self.direct_children:
            yield from c

    def printout(self, indentation=""):
        print(indentation, self.relative_prefix)
        if not self.value is None:
            print(indentation + "    " + self.value.name + " " + str(self.value.shape))
        for c in self.direct_children:
            c.printout(indentation + "    ")

def build(values, value_names_without_prefix=None, parent_prefix="", depth=0):
    if value_names_without_prefix is None:
        value_names_without_prefix = [v.name for v in values]
    assert len(values) == len(value_names_without_prefix)
    assert len(values) > 0
    values = list(values)
    value_names_without_prefix = list(value_names_without_prefix)

    prefix = get_common_prefix(value_names_without_prefix)

    node_value = None
    children = {}
    for value, value_name_without_prefix in zip(values, value_names_without_prefix):
        if len(value_name_without_prefix) == len(prefix):
            assert node_value is None
            node_value = value
        else:
            c = value_name_without_prefix[len(prefix)]
            if not c in children:
                children[c] = ([], [])
            children[c][0].append(value)
            children[c][1].append(value_name_without_prefix[len(prefix):])
    children = [build(values, value_names_without_prefix, parent_prefix=parent_prefix + prefix, depth=depth + 1) for values, value_names_without_prefix in children.values()]

    result = Node(
        value=node_value,
        relative_prefix=prefix,
        full_prefix=parent_prefix + prefix,
        direct_children=children,
        depth=depth,
    )
    for n in children:
        n.parent = result

    return result


import types
import numpy as np
import weightbridge.tree as tree

class Match:
    def __init__(self, in_nodes, out_nodes, description=None):
        assert isinstance(in_nodes, list) and isinstance(out_nodes, list)
        for n in in_nodes + out_nodes:
            assert isinstance(n, tree.Node)
        assert len(in_nodes) == len(set(id(n) for n in in_nodes))
        assert len(out_nodes) == len(set(id(n) for n in out_nodes))
        self.in_nodes = list(in_nodes)
        self.out_nodes = list(out_nodes)
        self.description = description

    def __sub__(self, other):
        assert isinstance(other, Match)
        return Match(
            [n for n in self.in_nodes if n not in other.in_nodes],
            [n for n in self.out_nodes if n not in other.out_nodes],
        )

format_map = {
    "pytorch": "pytorch",
    "tensorflow": "tensorflow",
    "tf": "tensorflow",
    "flax": "tensorflow",
    "haiku": "tensorflow",
    "jittor": "pytorch",
}

def can_reshape(in_shape, out_shape):
    if tuple(s for s in in_shape if s != 1) == tuple(s for s in out_shape if s != 1):
        return True
    in_shape = tuple(s for s in in_shape if s != 1)
    out_shape = tuple(s for s in out_shape if s != 1)
    if len(in_shape) == len(out_shape):
        return False
    shorter_shape = in_shape if len(in_shape) < len(out_shape) else out_shape
    longer_shape = in_shape if len(in_shape) > len(out_shape) else out_shape
    while len(longer_shape) > 0:
        if len(shorter_shape) == 0:
            return False
        elif longer_shape[0] == shorter_shape[0]:
            longer_shape = longer_shape[1:]
            shorter_shape = shorter_shape[1:]
        else:
            longer_shape = (longer_shape[0] * longer_shape[1],) + longer_shape[2:]
    if len(shorter_shape) > 0:
        return False
    return True

class State:
    def __init__(self, in_values, out_values, in_separator, out_separator, in_format, out_format, hints=[], verbose=False):
        if (in_format is None) != (out_format is None):
            raise ValueError("Either both or neither of in_format and out_format must be specified")
        if not in_format is None:
            if not in_format in format_map:
                raise ValueError(f"Format {in_format} not supported")
            if not out_format in format_map:
                raise ValueError(f"Format {out_format} not supported")
            self.in_format = format_map[in_format]
            self.out_format = format_map[out_format]
        else:
            self.in_format = self.out_format = None
        self.in_separator = in_separator
        self.out_separator = out_separator
        self.verbose = verbose
        self.hints = hints

        self.in_values = [types.SimpleNamespace(
                name=name,
                value=in_value,
                shape=tuple(in_value.shape),
                other=None,
            ) for name, in_value in in_values.items()]
        self.out_values = [types.SimpleNamespace(
                name=name,
                shape=tuple(out_value.shape),
                other=None,
            ) for name, out_value in out_values.items()]

        self.paired_nodes = []

    def pair_node(self, in_node, out_node):
        assert out_node.is_leaf() == in_node.is_leaf()
        assert out_node.other is None and in_node.other is None

        out_node.other = in_node
        in_node.other = out_node

        if out_node.is_leaf():
            out_node.value.other = in_node.value
            in_node.value.other = out_node.value

        self.paired_nodes.append((out_node, in_node))

    def get_format_permutation(self, out_shape, in_name, out_name):
        if self.in_format is not None and self.in_format != self.out_format:
            if self.in_format == "pytorch" and self.out_format == "tensorflow":
                if (in_name.endswith(".weight") or in_name == "weight") and len(out_shape) >= 2 or (in_name.endswith(".in_proj_weight") and len(out_shape) == 2):
                    return list(range(2, len(out_shape))) + [1, 0]
            elif self.in_format == "tensorflow" and self.out_format == "pytorch":
                print("weightbridge - warning: Tensorflow to PyTorch conversion is not tested") # TODO: test this
                if (out_name.endswith(".weight") or out_name == "weight") and len(out_shape) >= 2 or (out_name.endswith(".in_proj_weight") and len(out_shape) == 2):
                    return [len(out_shape) - 1, len(out_shape) - 2] + list(range(len(out_shape) - 2))
            else:
                raise ValueError(f"Conversion from {self.in_format} to {self.out_format} not supported")
        return None

    def adapt_format(self, value, out_shape, in_name, out_name):
        perm = self.get_format_permutation(out_shape, in_name, out_name)
        if not perm is None:
            perm_inv = [perm.index(i) for i in range(len(perm))]
            in_shape = [out_shape[i] for i in perm_inv]

            if value.shape != in_shape:
                assert can_reshape(value.shape, in_shape), f"Cannot reshape from {value.shape} to {in_shape} for {in_name} -> {out_name}"
                value = np.reshape(value, in_shape)
            value = np.transpose(value, perm)
            assert value.shape == out_shape, f"{value.shape} != {out_shape}"

        if value.shape != out_shape:
            assert can_reshape(value.shape, out_shape), f"Cannot reshape from {value.shape} to {out_shape} for {in_name} -> {out_name}"
            value = np.reshape(value, out_shape)

        return value

def matches_to_id(matches):
    return sorted(
        [(set(n.full_prefix for n in match.in_nodes), set(n.full_prefix for n in match.out_nodes)) for match in matches],
        key=lambda m: len(m[0]) + len(m[1]),
    )

def matcher(func):
    def wrapped(state, matches_in, *args, **kwargs):
        all_matches_out = []
        for match_in in matches_in:
            assert isinstance(match_in, Match)
            matches_out = list(func(state, match_in, *args, **kwargs))
            for match_out in matches_out:
                assert isinstance(match_out, Match)

            # All remaining nodes into single match
            match_remaining = match_in
            for match_out in matches_out:
                match_remaining = match_remaining - match_out
            matches_out.append(match_remaining)

            # Remove empty or unmatchable pairs
            i = 0
            while i < len(matches_out):
                if len(matches_out[i].in_nodes) == 0 or len(matches_out[i].out_nodes) == 0:
                    if not all(not n.is_leaf() for n in matches_out[i].in_nodes) or not all(not n.is_leaf() for n in matches_out[i].out_nodes):
                        def check(nodes, type):
                            if any(n.is_leaf() for n in nodes):
                                print(f"Matcher {func.__name__} yielded unmatched leafs:")
                                for n in nodes:
                                    if n.is_leaf():
                                        print(f"    {type} {n.full_prefix} {n.get_structured_shapes()}")
                        check(matches_out[i].in_nodes, "IN ")
                        check(matches_out[i].out_nodes, "OUT")
                        print()
                        print("Got OUT values:")
                        for v in state.out_values:
                            print(f"    OUT {v.name} {v.shape}")
                        print()
                        print("Got IN  values:")
                        for v in state.in_values:
                            print(f"    IN  {v.name} {v.shape}")
                        raise ValueError("Matcher yielded unmatched leafs")
                    del matches_out[i]
                else:
                    i += 1

            for match_out in matches_out:
                if len(match_out.in_nodes) == 1 and len(match_out.out_nodes) == 1:
                    state.pair_node(match_out.in_nodes[0], match_out.out_nodes[0])
                elif len(match_out.in_nodes) > 0 or len(match_out.out_nodes) > 0:
                    all_matches_out.append(match_out)
                else: # Empty match
                    pass

                if state.verbose and not match_out.description is None and (matches_to_id([match_in]) != matches_to_id(matches_out)):
                    print(match_out.description)
                    for n in match_out.out_nodes:
                        print(f"    OUT {n.full_prefix} {n.get_structured_shapes()}")
                    for n in match_out.in_nodes:
                        print(f"    IN  {n.full_prefix} {n.get_structured_shapes()}")

        changed = matches_to_id(matches_in) != matches_to_id(all_matches_out)
        return all_matches_out, changed
    wrapped.__name__ = func.__name__
    return wrapped

def hint_matcher(func):
    def wrapped(state, match):
        hints, is_match, hint_description = func(state, match)

        for hint in hints:
            # Build score matrix for all pairs of nodes
            matrix = np.zeros((len(match.out_nodes), len(match.in_nodes)), dtype="int32")
            for out_index in range(len(match.out_nodes)):
                for in_index in range(len(match.in_nodes)):
                    matrix[out_index, in_index] = 1 if is_match(match.out_nodes[out_index], match.in_nodes[in_index], hint) else 0
            s = np.sum(matrix)
            if s < len(match.out_nodes) or s == len(match.out_nodes) * len(match.in_nodes):
                continue

            # Find contiguous groups of matched nodes in matrix
            index_groups = set()
            checked_pairs = set()
            for out_index in range(len(match.out_nodes)):
                for in_index in range(len(match.in_nodes)):
                    if not (out_index, in_index) in checked_pairs:
                        if matrix[out_index, in_index] == 1:
                            out_indices = set()
                            in_indices = set()
                            out_indices.add(out_index)
                            in_indices.add(in_index)
                            for out_index2 in range(len(match.out_nodes)):
                                if matrix[out_index2, in_index] == 1:
                                    out_indices.add(out_index2)
                                    checked_pairs.add((out_index2, in_index))
                            for in_index2 in range(len(match.in_nodes)):
                                if matrix[out_index, in_index2] == 1:
                                    in_indices.add(in_index2)
                                    checked_pairs.add((out_index, in_index2))

                            out_indices = tuple(sorted(out_indices))
                            in_indices = tuple(sorted(in_indices))
                            index_groups.add((out_indices, in_indices))
                        checked_pairs.add((out_index, in_index))

            # Merge all groups with unequal number of nodes
            merged_out_indices = set()
            merged_in_indices = set()
            index_groups_out = set()
            for out_indices, in_indices in index_groups:
                if len(out_indices) == len(in_indices):
                    index_groups_out.add((out_indices, in_indices))
                else:
                    merged_out_indices.update(out_indices)
                    merged_in_indices.update(in_indices)
            if len(merged_out_indices) != len(merged_in_indices):
                continue
            if len(merged_out_indices) > 0:
                out_indices = tuple(sorted(merged_out_indices))
                in_indices = tuple(sorted(merged_in_indices))
                index_groups_out.add((out_indices, in_indices))
            index_groups = index_groups_out

            # End if more than one group was found
            if len(index_groups) > 1:
                remaining_out_nodes = list(match.out_nodes)
                remaining_in_nodes = list(match.in_nodes)
                for out_indices, in_indices in index_groups:
                    out_group = [match.out_nodes[i] for i in out_indices]
                    in_group = [match.in_nodes[i] for i in in_indices]
                    for n in out_group:
                        assert n in remaining_out_nodes
                        remaining_out_nodes.remove(n)
                    for n in in_group:
                        assert n in remaining_in_nodes
                        remaining_in_nodes.remove(n)
                    yield Match(in_group, out_group, f"Found subgroup resulting from {hint_description(hint)}")
                yield Match(remaining_in_nodes, remaining_out_nodes)
                return

        yield match
    wrapped.__name__ = func.__name__
    wrapped = matcher(wrapped)
    return wrapped


from .base import matcher, Match

@matcher
def match_by_paired_children(state, match):
    out_nodes = list(match.out_nodes)
    in_nodes = list(match.in_nodes)
    for pair_out_node, pair_in_node in state.paired_nodes:
        out_matches = [n for n in out_nodes if pair_out_node.full_prefix.startswith(n.full_prefix)]
        in_matches = [n for n in in_nodes if pair_in_node.full_prefix.startswith(n.full_prefix)]
        if len(out_matches) == 1 and len(in_matches) == 1:
            out_node = out_matches[0]
            in_node = in_matches[0]
            yield Match([in_node], [out_node], f"Matched by paired children of OUT {pair_out_node.full_prefix} IN {pair_in_node.full_prefix}")
            out_nodes.remove(out_node)
            in_nodes.remove(in_node)


from collections import defaultdict
import numpy as np
from .base import matcher, Match
from .util import extract_number_expression
import re

@matcher
def match_number_regex(state, match):
    def make_groups(nodes):
        nodes = [n for n in nodes]
        groupswithmatchingregex = []
        while len(nodes) > 1:
            # Find possible prefixes
            prefix_expressions = set()
            for expr, _ in set(extract_number_expression(n.full_prefix) for n in nodes):
                prefix = ""
                for t in expr.split("([0-9]+)")[:-1]:
                    prefix = prefix + t + "([0-9]+)"
                    prefix_expressions.add(prefix)
            prefix_expressions = sorted(prefix_expressions, key=lambda expr: len(expr))

            for prefix_expr in prefix_expressions:
                prefix_expr = re.compile(prefix_expr)
                matches = [prefix_expr.match(n.full_prefix) for n in nodes]
                if sum((1 if m else 0) for m in matches) > 1:
                    matching_nodes = [n for n, m in zip(nodes, matches) if m]
                    matches = [m for m in matches if m]

                    # Find the first varying number in the list of numbers per node
                    numbers = np.asarray([[int(g) for g in match.groups()] for match in matches]) # matches nums
                    unique_nums = np.asarray([len(set(numbers[:, num_index])) for num_index in range(numbers.shape[1])]) # nums
                    for used_num_index in range(len(unique_nums)):
                        if unique_nums[used_num_index] > 1:
                            break
                    else:
                        continue
                    numbers = numbers[:, used_num_index] # matches

                    groupswithmatchingregex.append([(n, m) for n, m in zip(numbers, matching_nodes)])
                    for n in matching_nodes:
                        nodes.remove(n)
                    break # Found a prefix that counts more than one node
            else:
                break # No prefix matches

        for n in nodes:
            groupswithmatchingregex.append([(None, n)])

        num_to_groupswithmatchingregex = defaultdict(list)
        for g in groupswithmatchingregex:
            num_to_groupswithmatchingregex[len(g)].append(g)

        return num_to_groupswithmatchingregex

    # Divide nodes into groups that look like: prefix number postfix
    out_num_to_groupswithmatchingregex = make_groups(match.out_nodes)
    in_num_to_groupswithmatchingregex = make_groups(match.in_nodes)

    # Pair groups that have the same unique size
    unique_nums = set(out_num_to_groupswithmatchingregex.keys()).intersection(set(in_num_to_groupswithmatchingregex.keys()))
    for k in unique_nums:
        out_groups = out_num_to_groupswithmatchingregex[k]
        in_groups = in_num_to_groupswithmatchingregex[k]
        if len(out_groups) == 1 and len(in_groups) == 1:
            # Unique number of nodes -> divide by parsed number values
            def to_dict(group):
                nums_to_nodes = defaultdict(list)
                for num, node in group:
                    nums_to_nodes[num].append(node)
                return nums_to_nodes
            out_nums_to_nodes = to_dict(out_groups[0])
            in_nums_to_nodes = to_dict(in_groups[0])
            if len(out_nums_to_nodes) == len(in_nums_to_nodes):
                # Sort by parsed number values
                out_sorted_groups = [nodes for num, nodes in sorted(out_nums_to_nodes.items(), key=lambda t: t[0])]
                in_sorted_groups = [nodes for num, nodes in sorted(in_nums_to_nodes.items(), key=lambda t: t[0])]
                assert len(out_sorted_groups) == len(in_sorted_groups)

                # Match
                for out_group, in_group in zip(out_sorted_groups, in_sorted_groups):
                    yield Match(in_group, out_group, "Found counted modules")


from .base import hint_matcher

@hint_matcher
def match_equivalent_hardcoded_leafs(state, match):
    hints = [[
        ["weight", "scale", "gamma", "w"],
        ["bias", "offset", "beta", "b"],
        ["moving_mean", "running_mean", "~/mean_ema/average", "mean_ema/average", "mean/average"],
        ["moving_variance", "running_var", "~/var_ema/average", "var_ema/average", "var/average"],
        ["key", "keys"],
        ["query", "queries"],
    ]]

    out_nodes = [n for n in match.out_nodes if n.is_leaf()]
    in_nodes = [n for n in match.in_nodes if n.is_leaf()]

    def get_index(node, separator, hint):
        for i, equivalent_postfixes in enumerate(hint):
            if any(node.full_prefix.endswith(separator + postfix) for postfix in equivalent_postfixes):
                return i
        return -1

    def is_valid_hint(hint):
        return all(get_index(n, state.out_separator, hint) >= 0 for n in out_nodes) and all(get_index(n, state.in_separator, hint) >= 0 for n in in_nodes)

    hints = [h for h in hints if is_valid_hint(h)]

    def is_match(out_node, in_node, hint):
        return get_index(out_node, state.out_separator, hint) == get_index(in_node, state.in_separator, hint)

    return hints, is_match, lambda hint: f"equivalent hardcoded leafs"


from .base import matcher, Match
from .util import extract_number_expression
import re

@matcher
def match_known_paired_number_regex(state, match):
    last_pairs_num_name = "match_known_paired_number_regex_last_pairs_num"
    if not last_pairs_num_name in vars(state):
        vars(state)[last_pairs_num_name] = 0
    hints_name = "match_known_paired_number_regex_hints"
    if not hints_name in vars(state):
        vars(state)[hints_name] = {}
    hints = vars(state)[hints_name]

    for out_node, in_node in state.paired_nodes[vars(state)[last_pairs_num_name]:]:
        out_expr, out_nums = extract_number_expression(out_node.full_prefix)
        in_expr, in_nums = extract_number_expression(in_node.full_prefix)
        if out_nums > 0 and in_nums > 0:
            hint = (out_expr, in_expr)
            if not hint in hints:
                hints[hint] = (re.compile(hint[0]), re.compile(hint[1]))
    vars(state)[last_pairs_num_name] = len(state.paired_nodes)

    def is_match(name, expr):
        re_match = re.match(expr, name)
        if re_match:
            return tuple(int(g) for g in re_match.groups())
        else:
            return None

    out_nodes = list(match.out_nodes)
    in_nodes = list(match.in_nodes)
    for hint in hints.values():
        def get_matches(nodes, hint):
            matches = [(n, is_match(n.full_prefix, hint)) for n in nodes]
            matches = [(n, nums) for n, nums in matches if not nums is None]
            matches = sorted(matches, key=lambda t: t[1])
            nums = [num for _, nums in matches for num in nums]
            if len(nums) != len(set(nums)):
                return None
            matches = [n for n, nums in matches]
            return matches
        out_matching_nodes = get_matches(out_nodes, hint[0])
        in_matching_nodes = get_matches(in_nodes, hint[1])
        if not out_matching_nodes is None and not in_matching_nodes is None and len(out_matching_nodes) == len(in_matching_nodes) and len(out_matching_nodes) > 0:
            for out_node, in_node in zip(out_matching_nodes, in_matching_nodes):
                yield Match([in_node], [out_node], f"Found counted nodes from paired hint OUT {hint[0].pattern[1:-1]} IN  {hint[1].pattern[1:-1]}:")
                out_nodes.remove(out_node)
                in_nodes.remove(in_node)



from .base import matcher, Match

@matcher
def match_by_paired_parents(state, match):
    def get_first_paired_parent(node):
        if not node.other is None:
            return node
        elif node.parent is None:
            return None
        else:
            return get_first_paired_parent(node.parent)

    def to_dict(nodes):
        pairedparent_to_nodes = {}
        for n in nodes:
            parent = get_first_paired_parent(n)
            if not parent is None:
                if not parent.full_prefix in pairedparent_to_nodes:
                    pairedparent_to_nodes[parent.full_prefix] = (parent, [])
                pairedparent_to_nodes[parent.full_prefix][1].append(n)
        return pairedparent_to_nodes

    # Link every paired parent to nodes in the current match-group
    out_pairedparent_to_nodes = to_dict(match.out_nodes)
    in_pairedparent_to_nodes = to_dict(match.in_nodes)

    for out_parent, out_children in out_pairedparent_to_nodes.values():
        in_parent = out_parent.other
        in_children = in_pairedparent_to_nodes[in_parent.full_prefix][1]
        yield Match(in_children, out_children, f"Found subtrees with matched parents OUT {out_parent.full_prefix} IN  {in_parent.full_prefix}")


from collections import defaultdict
from .base import matcher, Match

def structured_shapes_depth(shape):
    if isinstance(shape, tuple) and all(isinstance(s, int) for s in shape):
        return 1
    else:
        assert all(not isinstance(s, int) for s in shape)
        return 1 + max(structured_shapes_depth(s) for s in shape)

@matcher
def match_by_structured_shapes(state, match, **kwargs):
    pairs_to_id = lambda nodes: [n.full_prefix for n in nodes]

    def to_uniquekey_dict(nodes):
        result = defaultdict(list)
        for n in nodes:
            if n.other is None:
                result[n.get_structured_shapes(**kwargs)].append(n)
        return result
    out_uniquekey_to_nodes = to_uniquekey_dict(match.out_nodes)
    in_uniquekey_to_nodes = to_uniquekey_dict(match.in_nodes)

    intersection_keys = set(out_uniquekey_to_nodes.keys()).intersection(in_uniquekey_to_nodes.keys())
    intersection_keys = sorted(intersection_keys, key=lambda shapes: -structured_shapes_depth(shapes))

    for key in intersection_keys:
        yield Match(in_uniquekey_to_nodes[key], out_uniquekey_to_nodes[key], f"Found matches with structured shapes {key}")

    for key in out_uniquekey_to_nodes.keys():
        if key not in intersection_keys:
            yield Match([], out_uniquekey_to_nodes[key])
    for key in in_uniquekey_to_nodes.keys():
        if key not in intersection_keys:
            yield Match(in_uniquekey_to_nodes[key], [])


#!/usr/bin/env python3

from setuptools import setup, find_packages

setup(
    name="weightbridge",
    version="0.0.1",
    python_requires=">=3",
    description="Lightweight library to map model weights between different model implementations",
    author="Florian Fervers",
    author_email="florian.fervers@gmail.com",
    url="https://github.com/fferflo/weightbridge",
    packages=find_packages(),
    license="MIT",
    include_package_data=True,
    install_requires=[
        "numpy",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
)
